import myers
from copy import copy

from .grouped_shader import CodeLineGroup, DiffedShaderWithGroupedLines
from .type_aliases import FunctionName, ShaderFlags, ShaderLineIndex
from .permutation import EncodedUniqufiedPermutations


class DiffedCode:
    lines: list[ShaderLineIndex]
    line_conditions: list[list[ShaderFlags]]

    def __init__(self):
        self.lines = []
        self.line_conditions = []

    def group_lines(self):
        """
        Groups consecutive lines with identical conditions into line group objects
        """
        line_groups: list[CodeLineGroup] = []

        if len(self.lines) == 0:
            return line_groups

        current_group = CodeLineGroup()
        current_group.condition = self.line_conditions[0]
        for line, condition in zip(self.lines, self.line_conditions):
            if condition != current_group.condition:
                line_groups.append(current_group)
                current_group = CodeLineGroup()
                current_group.condition = condition

            current_group.lines.append(line)

        line_groups.append(current_group)
        return line_groups


class DiffedShader:
    main_code: DiffedCode
    functions: dict[FunctionName, DiffedCode]

    def __init__(self):
        self.main_code = DiffedCode()
        self.functions = {}

    def group_lines(self):
        grouped_shader = DiffedShaderWithGroupedLines()

        grouped_shader.main_code = self.main_code.group_lines()
        for func_name, func in self.functions.items():
            grouped_shader.functions[func_name] = func.group_lines()

        return grouped_shader


def diff_permutations(encoded_permutations: EncodedUniqufiedPermutations):
    """
    Combines (by diffing) permutations of code together into one `DiffedCode` object
    """
    lines: list[ShaderLineIndex] = []
    new_conditions: list[list[ShaderFlags]]
    line_conditions: list[list[ShaderFlags]] = []
    for code, flag_list in zip(encoded_permutations.codes, encoded_permutations.flags):
        diff = myers.diff(lines, code)
        lines = []
        new_conditions = []
        current_index = 0
        for op, val in diff:
            lines.append(val)
            if op == "i":
                new_conditions.append(copy(flag_list))
            elif op == "r":
                new_conditions.append(line_conditions[current_index])
                current_index += 1
            elif op == "k":
                condition = line_conditions[current_index]
                condition.extend(flag_list)
                new_conditions.append(condition)
                current_index += 1
        line_conditions = new_conditions

    diffed_shader = DiffedCode()
    diffed_shader.lines = lines
    diffed_shader.line_conditions = line_conditions

    return diffed_shader



"This file contains commonly used type aliases"

FlagName = str
FlagValue = str
ShaderFlags = dict[FlagName, FlagValue]

FlagDefinition = dict[FlagName, list[FlagValue]]
"A list of all flag values for a given flag name"


ShaderCode = str
FunctionName = str


ShaderLineIndex = int
"Unique index of a unique line of code"

ShaderLine = str
"A single line of code"



import subprocess

from lazurite.material.stage import ShaderStage
from lazurite.material.platform import ShaderPlatform


class Glslang:
    glslang_path: str

    def __init__(self, glslang_paths: list[str] | str | None = None) -> None:
        if glslang_paths is None:
            glslang_paths = ["./glslang", "glslang"]
        elif isinstance(glslang_paths, str):
            glslang_paths = [glslang_paths]

        self.glslang_path = ""

        for path in glslang_paths:
            try:
                result = subprocess.run(
                    [path, "--stdin", "-S", "frag"],
                    capture_output=True,
                    input=b"",
                )
            except FileNotFoundError:
                pass
            else:
                self.glslang_path = path
                break

        if not self.glslang_path:
            raise Exception(
                f"Error! No valid GLSLANG validator was found in the list {glslang_paths}"
            )

        if result.returncode:
            print(result.stdout.decode())
            print(result.stderr.decode())
            result.check_returncode()

    def validate(
        self,
        code: str | bytes,
        platform: ShaderPlatform = ShaderPlatform.ESSL_310,
        stage: ShaderStage = ShaderStage.Fragment,
    ):
        if not platform.name.startswith(("GLSL", "ESSL")):
            raise Exception(
                f'Invalid platform "{platform.name}", only GLSL and ESSL platforms support glslang validation.'
            )

        args = [self.glslang_path, "--stdin"]

        args.append("-S")
        if stage == ShaderStage.Compute:
            args.append("comp")
        elif stage == ShaderStage.Vertex:
            args.append("vert")
        else:
            args.append("frag")

        version_string = platform.name[-3:]
        if platform in (ShaderPlatform.ESSL_300, ShaderPlatform.ESSL_310):
            version_string += "es"
        args.extend(("--glsl-version", version_string))

        if isinstance(code, str):
            code = code.encode()

        result = subprocess.run(args, capture_output=True, input=code)

        log = []
        if result.stdout:
            log.append(result.stdout.decode())
        if result.stderr:
            log.append(result.stderr.decode())

        has_log = bool(log)
        log = "\n\n".join([""] + log + ["Command: " + " ".join(args)])

        if result.returncode:
            raise Exception(log)

        return log if has_log else ""



import time
from enum import Enum, auto
from copy import copy

from .type_aliases import ShaderFlags, FlagDefinition, FlagName, FlagValue
from .grouped_shader import DiffedShaderWithGroupedLines, CodeLineGroup
from .local_flag_definition import LocalFlagDeinition
from .all_flags import AllFlags

FlagsOutcome = bool
"Determines whether expression should be true for a given set of flags"


class ExpressionSearchInput:
    """
    Input object for expression search algorithms
    """

    flags: list[tuple[FlagsOutcome, ShaderFlags]]
    flag_definition: FlagDefinition

    def __init__(self):
        self.flags = []
        self.flag_definition = {}

    def __eq__(self, value):
        if not isinstance(value, type(self)):
            return False
        return (
            self.flags == value.flags and self.flag_definition == value.flag_definition
        )

    @classmethod
    def from_diffed_grouped_shader(
        cls,
        shader: DiffedShaderWithGroupedLines,
        flag_def: LocalFlagDeinition,
        all_flags: AllFlags,
    ):
        calc_list: list[ExpressionSearchInput] = []
        cls._extract_search_inputs(
            shader.main_code, all_flags.main_flags, flag_def.main_shader, calc_list
        )

        for func_name, func_body in shader.functions.items():
            cls._extract_search_inputs(
                func_body,
                all_flags.function_flags[func_name],
                flag_def.functions[func_name],
                calc_list,
            )

        return calc_list

    @classmethod
    def _extract_search_inputs(
        cls,
        code_line_groups: list[CodeLineGroup],
        all_flags: list[ShaderFlags],
        flag_def: FlagDefinition,
        expr_search_input_list: list["ExpressionSearchInput"],
    ):
        for line_group in code_line_groups:
            if len(line_group.condition) == len(all_flags):
                continue
            search_input = cls()
            search_input.flag_definition = flag_def
            search_input.flags = [
                (flags in line_group.condition, flags) for flags in all_flags
            ]
            try:
                index = expr_search_input_list.index(search_input)
            except ValueError:
                index = len(expr_search_input_list)
                expr_search_input_list.append(search_input)
            line_group.expression_search_index = index


class JoinType(Enum):
    """
    Determines how expression search token should be combined with other tokens.

    - `Or` - Token is boolean ORed with the result of evaluating previous tokens
    - `And` - Token is boolean ANDed with the result of evaluating previous tokens
    - `Initial` - Current expression value is directly equal to the token value. This must only be used for the first token in the sequence

    For example, the string of tokens `[token0, token1, token2, token3]` could
    form the following expression `(token0 && token1) || token2) && token3`
    in which case tokens would have the following join types:
    - token0 - `Initial` (the first token must always have `Initial` join type)
    - token1 - `And`
    - token2 - `Or`
    - token3 - `And`
    """

    Or = auto()
    And = auto()
    Initial = auto()


class ExpressionSearchToken:
    """
    Intermediate object used in expression search algorithm.
    """

    join_type: JoinType
    "Controls how this token is combined with previous tokens in a boolean expression"
    is_negative: bool
    "Determines whether the token is negated or not"
    flag_name: FlagName
    flag_value: FlagValue

    def __init__(self):
        self.join_type = JoinType.Initial
        self.is_negative = False
        self.flag_name = ""
        self.flag_value = ""

    def __copy__(self):
        obj = ExpressionSearchToken()

        obj.join_type = self.join_type
        obj.is_negative = self.is_negative
        obj.flag_name = self.flag_name
        obj.flag_value = self.flag_value

        return obj

    def __deepcopy__(self, memo):
        obj = ExpressionSearchToken()

        obj.join_type = self.join_type
        obj.is_negative = self.is_negative
        obj.flag_name = self.flag_name
        obj.flag_value = self.flag_value

        return obj


class ExpressionSearchOutput:
    """
    Object representing the results of expression search algorithm - it holds
    the best found expression token sequence, along with its score
    """

    token_list: list[ExpressionSearchToken]
    score: int

    def __init__(self):
        self.token_list = []
        self.score = 0


def _evaluate_expression(
    expression_token_list: list[ExpressionSearchToken], flags: ShaderFlags
) -> bool:
    """
    Evaluates expression token sequence for a given set of flags.
    """
    # Evaluation is in reverse order, as that allows to perform short-circuiting
    for token in reversed(expression_token_list):
        token_value = flags.get(token.flag_name, None) == token.flag_value

        if token.is_negative:
            token_value = not token_value

        if token.join_type is JoinType.And:
            if not token_value:
                return False
        elif token.join_type is JoinType.Or:
            if token_value:
                return True
        elif token.join_type is JoinType.Initial:
            return token_value


def _calc_score(
    expression_token_list: list[ExpressionSearchToken],
    flag_outcomes: list[tuple[FlagsOutcome, ShaderFlags]],
):
    """
    Calculates the score of a sequence, which is equal to the number of sets of flags that it can correctly match
    """
    score = 0
    for outcome, flags in flag_outcomes:
        score += (
            1 if _evaluate_expression(expression_token_list, flags) == outcome else 0
        )

    return score


def _fast_search(input: ExpressionSearchInput):
    """
    This algorithm attempts to find a sequence of expression tokens that correctly executes for all sets of flags.
    It works by testing one token at a time - finding a configuration of a token that yields the highest score, then
    moving on to the next token and doing the same.

    This function is fast because it has linear complexity in the token sequence length, but it's not guaranteed to find an exact solution.
    """
    best_expression: list[ExpressionSearchToken] = []
    best_expression_score = 0

    current_expression: list[ExpressionSearchToken] = []

    for _ in range(len(input.flag_definition) + 5):
        best_token = ExpressionSearchToken()
        best_token_score = 0
        current_expression.append(copy(best_token))

        join_list = (
            (JoinType.Initial,)
            if len(current_expression) == 1
            else (JoinType.Or, JoinType.And)
        )
        token = current_expression[-1]
        for is_negative in (False, True):
            token.is_negative = is_negative
            for join_type in join_list:
                token.join_type = join_type
                for flag_name, flag_values in input.flag_definition.items():
                    token.flag_name = flag_name
                    for flag_value in flag_values:
                        token.flag_value = flag_value

                        score = _calc_score(current_expression, input.flags)

                        if score > best_token_score:
                            best_token_score = score
                            best_token = copy(token)

        current_expression[-1] = best_token

        if best_token_score > best_expression_score:
            best_expression = current_expression[:]
            best_expression_score = best_token_score

        if best_expression_score >= len(input.flags):
            break

    return best_expression_score, best_expression


def _increment_expression(
    expression: list[ExpressionSearchToken], flag_def: FlagDefinition
):
    """
    This function cycles through all possible sequences of tokens.
    Given a sequence of tokens, it creates the next sequence by changing some of the properties of tokens or appending a new token at the end.
    """
    for token in expression:
        # Increment flag value
        flag_value_list = flag_def[token.flag_name]
        new_value_index = flag_value_list.index(token.flag_value) + 1

        if new_value_index != len(flag_value_list):
            token.flag_value = flag_value_list[new_value_index]
            return

        # Increment flag name
        flag_name_list = list(flag_def.keys())
        new_name_index = flag_name_list.index(token.flag_name) + 1

        if new_name_index != len(flag_name_list):
            token.flag_name = flag_name_list[new_name_index]
            token.flag_value = flag_def[token.flag_name][0]
            return

        token.flag_name = flag_name_list[0]
        token.flag_value = flag_def[token.flag_name][0]

        # Increment join type
        if token.join_type != JoinType.Initial:
            if token.join_type == JoinType.Or:
                token.join_type = JoinType.And
                return
            token.join_type = JoinType.Or

        # Increment is_negative
        if not token.is_negative:
            token.is_negative = True
            return
        token.is_negative = False

    # If all values of all tokens were reset, add a new token at the end
    token = ExpressionSearchToken()
    token.flag_name = list(flag_def.keys())[0]
    token.flag_value = flag_def[token.flag_name][0]
    token.is_negative = False
    token.join_type = JoinType.Initial if len(expression) == 0 else JoinType.Or

    expression.append(token)


def _slow_search(input: ExpressionSearchInput, timeout: float = 10):
    """
    This algorithm attempts to find a sequence of expression tokens that correctly executes for all sets of flags.

    It's a brute-force algorithm that checks every possible combination of tokens. Given infinite time it is guaranteed to find the exact solution,
    but because it has an exponential complexity in the number of tokens in a sequence, it can be quite slow, which is why a timeout parameter is necessary.
    """
    best_expression: list[ExpressionSearchToken] = []
    best_expression_score = 0

    current_expression: list[ExpressionSearchToken] = []
    t = time.perf_counter()
    while True:
        score = _calc_score(current_expression, input.flags)

        if score > best_expression_score:
            best_expression_score = score
            best_expression = [copy(token) for token in current_expression]

        if (
            best_expression_score == len(input.flags)
            or time.perf_counter() - t >= timeout
        ):
            break

        _increment_expression(current_expression, input.flag_definition)

    return best_expression_score, best_expression


# This is a work-in-progress new search algorithm,
# although I might remove it because it kinda sucks in comparison to a combination of fast + slow search
def _hybrid_search(input: ExpressionSearchInput, timeout: float = 10):
    best_expression: list[ExpressionSearchToken] = []
    best_expression_score = 0

    INITIAL_TOKEN = ExpressionSearchToken()
    INITIAL_TOKEN.is_negative = False
    INITIAL_TOKEN.join_type = JoinType.Or
    INITIAL_TOKEN.flag_name = list(input.flag_definition.keys())[0]
    INITIAL_TOKEN.flag_value = input.flag_definition[INITIAL_TOKEN.flag_name][0]

    reached_timeout = False

    t = time.perf_counter()
    # for search_width in range(1, 100):
    for search_width in (1, 2, 4, 8, 16, 32):
        expression: list[ExpressionSearchToken] = []
        expression_score = 0

        while len(expression) < len(input.flag_definition) + 5:
            expression_extension: list[ExpressionSearchToken] = [copy(INITIAL_TOKEN)]
            best_extension: list[ExpressionSearchToken] = [copy(INITIAL_TOKEN)]
            best_extension_score = 0

            while True:
                score = _calc_score(expression + expression_extension, input.flags)

                if score == len(input.flags):
                    return score, expression + expression_extension

                if score > best_extension_score:
                    best_extension_score = score
                    best_extension = [copy(token) for token in expression_extension]

                _increment_expression(expression_extension, input.flag_definition)

                reached_timeout = time.perf_counter() - t >= timeout
                if len(expression_extension) > search_width or reached_timeout:
                    break

            expression = expression + best_extension
            expression_score = best_extension_score

            if reached_timeout:
                break

        if expression_score > best_expression_score:
            best_expression_score = expression_score
            best_expression = expression

        if reached_timeout:
            break

    return best_expression_score, best_expression


def expression_search(inputs: list[ExpressionSearchInput], timeout: float = 10):
    """
    This function applies search algorithms in order to find a boolean expression that would correctly match all sets of flags.
    """
    output_list: list[ExpressionSearchOutput] = []
    for input in inputs:
        score, result = _fast_search(input)
        # score, result = _hybrid_search(input, timeout)

        if score != len(input.flags):
            print("Slow Search")

            slow_score, slow_result = _slow_search(input, timeout)

            if slow_score > score or (
                slow_score == score and len(slow_result) < len(result)
            ):
                score = slow_score
                result = slow_result

            if score < len(input.flags):
                print("Not Found")

        search_output = ExpressionSearchOutput()
        search_output.score = score
        search_output.token_list = result

        output_list.append(search_output)

    return output_list



from .type_aliases import FlagDefinition, FunctionName, FlagName
from .grouped_shader import DiffedShaderWithGroupedLines, CodeLineGroup


class LocalFlagDeinition:
    """
    Separate flag definitions for each context (main code and functions).
    """

    main_shader: FlagDefinition
    functions: dict[FunctionName, FlagDefinition]

    def __init__(self):
        self.main_shader = {}
        self.functions = {}

    @classmethod
    def from_diffed_grouped_shader(cls, shader: DiffedShaderWithGroupedLines):
        obj = cls()

        obj.main_shader = cls._flag_def_from_line_group_list(shader.main_code)
        for func_name, func in shader.functions.items():
            obj.functions[func_name] = cls._flag_def_from_line_group_list(func)

        return obj

    def _flag_def_from_line_group_list(line_list: list[CodeLineGroup]):
        flag_def: FlagDefinition = {}
        for line_group in line_list:
            for flags in line_group.condition:
                for key, value in flags.items():
                    value_list = flag_def.get(key, None)

                    if value_list is None:
                        value_list = []
                        flag_def[key] = value_list

                    if value not in value_list:
                        value_list.append(value)

        return flag_def

    def filter_and_bias_flags(self):
        """
        Removes flags that are always set and biases boolean flags (On/Off, Enabled/Disabled) in expression search towards enabling and against disabling
        """
        definitions: list[FlagDefinition] = list(self.functions.values())
        definitions.append(self.main_shader)

        for flag_def in definitions:
            keys_to_remove: list[FlagName] = []
            for flag_name, flag_values in flag_def.items():
                if len(flag_values) <= 1:
                    keys_to_remove.append(flag_name)
                    continue
                # Bias against disabling
                for i in ("Off", "Disabled"):
                    if i in flag_values:
                        flag_values.remove(i)
                        flag_values.append(i)
                # Bias towards enabling
                for i in ("On", "Enabled"):
                    if i in flag_values:
                        flag_values.remove(i)
                        flag_values.insert(0, i)

            for key in keys_to_remove:
                flag_def.pop(key)



from .diffing import DiffedShader, diff_permutations
from .permutation import (
    ShaderPermutation,
    PermutationBase,
    EncodedUniqufiedPermutations,
)
from .type_aliases import (
    ShaderCode,
    FunctionName,
    ShaderFlags,
    ShaderLine,
    ShaderLineIndex,
)


class EncodedShader:
    """
    This object represents a shader with each unique line of code replaced with a unique number
    """

    main_shader: EncodedUniqufiedPermutations
    functions: dict[FunctionName, EncodedUniqufiedPermutations]
    line_decode_table: list[ShaderLine]

    def __init__(self, permutations: list[ShaderPermutation]):
        self.main_shader = EncodedUniqufiedPermutations()
        self.line_decode_table = []
        self.functions = {}

        uniqufied_functions: dict[FunctionName, dict[ShaderCode, list[ShaderFlags]]] = (
            {}
        )
        uniqufied_main_shader: dict[ShaderCode, list[ShaderFlags]] = {}
        for permutation in permutations:
            # Uniquify main code
            self._insert_flags(uniqufied_main_shader, permutation)

            # Uniquify functions
            for name, func in permutation.functions.items():
                func_code_dict = uniqufied_functions.get(name, None)

                if func_code_dict is None:
                    func_code_dict = {}
                    uniqufied_functions[name] = func_code_dict

                self._insert_flags(func_code_dict, func)

        # Encode main code
        self._encode(uniqufied_main_shader, self.main_shader)

        # Encode functions
        for func_name, table in uniqufied_functions.items():
            encoded_func = EncodedUniqufiedPermutations()
            self._encode(table, encoded_func)
            self.functions[func_name] = encoded_func

    def _insert_flags(
        self, table: dict[ShaderCode, list[ShaderFlags]], permutation: PermutationBase
    ):
        flag_list = table.get(permutation.code, None)

        if flag_list is None:
            flag_list = []
            table[permutation.code] = flag_list

        flag_list.append(permutation.flags)

    def _encode(
        self,
        table: dict[ShaderCode, list[ShaderFlags]],
        shader: EncodedUniqufiedPermutations,
    ):
        for code, flag_list in table.items():
            encoded_shader: list[ShaderLineIndex] = []
            for line in code.splitlines():
                try:
                    line_index = self.line_decode_table.index(line)
                except ValueError:
                    line_index = len(self.line_decode_table)
                    self.line_decode_table.append(line)

                encoded_shader.append(line_index)

            shader.codes.append(encoded_shader)
            shader.flags.append(flag_list)

    def diff(self):
        """
        Diffs encoded shader, which combines together code of all permutations
        """
        diffed_shader = DiffedShader()
        diffed_shader.main_code = diff_permutations(self.main_shader)
        for func_name, func in self.functions.items():
            diffed_shader.functions[func_name] = diff_permutations(func)
        return diffed_shader



from enum import Enum, auto


class ShaderPlatform(Enum):
    Direct3D_SM40 = auto()
    Direct3D_SM50 = auto()
    Direct3D_SM60 = auto()
    Direct3D_SM65 = auto()
    Direct3D_XB1 = auto()
    Direct3D_XBX = auto()
    GLSL_120 = auto()
    GLSL_430 = auto()
    ESSL_300 = auto()  # Removed in format version 25
    ESSL_310 = auto()
    Metal = auto()
    Vulkan = auto()
    Nvn = auto()
    PSSL = auto()
    Unknown = auto()

    @classmethod
    def _platform_mapping(cls, version: int):
        if version >= 25:
            return {
                cls.Direct3D_SM40: 0,
                cls.Direct3D_SM50: 1,
                cls.Direct3D_SM60: 2,
                cls.Direct3D_SM65: 3,
                cls.Direct3D_XB1: 4,
                cls.Direct3D_XBX: 5,
                cls.GLSL_120: 6,
                cls.GLSL_430: 7,
                cls.ESSL_310: 8,
                cls.Metal: 9,
                cls.Vulkan: 10,
                cls.Nvn: 11,
                cls.PSSL: 12,
                cls.Unknown: 13,
                # Platform conversion
                cls.ESSL_300: cls.ESSL_310,
            }
        else:
            return {
                cls.Direct3D_SM40: 0,
                cls.Direct3D_SM50: 1,
                cls.Direct3D_SM60: 2,
                cls.Direct3D_SM65: 3,
                cls.Direct3D_XB1: 4,
                cls.Direct3D_XBX: 5,
                cls.GLSL_120: 6,
                cls.GLSL_430: 7,
                cls.ESSL_300: 8,
                cls.ESSL_310: 9,
                cls.Metal: 10,
                cls.Vulkan: 11,
                cls.Nvn: 12,
                cls.PSSL: 13,
                cls.Unknown: 14,
            }

    def get_value(self, version: int) -> int:
        mapping = self._platform_mapping(version)
        value = mapping.get(self, None)

        if isinstance(value, ShaderPlatform):
            value = mapping.get(value, None)

        if value is None:
            raise Exception(
                "Platform {self} is not supported in version {version} or there are no conversions available!"
            )

        return value

    def get_name(self, version: int) -> int:
        mapping = self._platform_mapping(version)
        platform = mapping.get(self, None)

        if isinstance(platform, int):
            platform = self
        elif platform is None:
            raise Exception(
                "Platform {self} is not supported in version {version} or there are no conversions available!"
            )

        return platform.name

    @classmethod
    def get_list(cls, version: int):
        platform_list: list[ShaderPlatform] = []

        for platform, value in cls._platform_mapping(version).items():
            if isinstance(value, int):
                platform_list.append(platform)

        return platform_list

    def file_extension(self):
        if self.name.startswith("Direct3D"):
            return "dxbc"

        elif self.name.startswith("GLSL") or self.name.startswith("ESSL"):
            return "glsl"

        elif self.name == "Metal":
            return "metal"

        elif self.name == "Vulkan":
            return "spirv"

        else:
            return "bin"



from .processing import format_function_name
from .type_aliases import ShaderLineIndex, ShaderFlags, FunctionName
from .all_flags import AllFlags


class CodeLineGroup:
    """
    A group of consecutive lines of code that share the same condition (appear when the same flags are set)
    """

    lines: list[ShaderLineIndex]
    condition: list[ShaderFlags]
    expression_search_index: int | None

    def __init__(self):
        self.lines = []
        self.condition = []
        self.expression_search_index = None

    def assemble(self, line_decode_table: list[str], macro_expressions: list[str]):
        code = "\n".join(line_decode_table[i] for i in self.lines)
        if self.expression_search_index is not None:
            code = f"{macro_expressions[self.expression_search_index]}\n{code}\n#endif"
        return code


class DiffedShaderWithGroupedLines:
    main_code: list[CodeLineGroup]
    functions: dict[FunctionName, list[CodeLineGroup]]

    def __init__(self):
        self.main_code = []
        self.functions = {}

    def gen_all_flags_list(self):
        """
        Generates a list of all flags, per context (main code or individual functions)
        """
        all_flags = AllFlags()

        all_flags.main_flags = self._gen_flag_list_from_line_groups(self.main_code)

        for function_name, function_line_groups in self.functions.items():
            all_flags.function_flags[function_name] = (
                self._gen_flag_list_from_line_groups(function_line_groups)
            )

        return all_flags

    @staticmethod
    def _gen_flag_list_from_line_groups(code_line_groups: list[CodeLineGroup]):
        flag_list: list[ShaderFlags] = []
        for line_group in code_line_groups:
            for flags in line_group.condition:
                if flags not in flag_list:
                    flag_list.append(flags)

        return flag_list

    def assemble_code(
        self,
        macro_expressions: list[str],
        line_decode_table: list[str],
    ):
        """
        Assembles shader back into its source code form
        """
        shader_code = self.assemble_line_groups(
            self.main_code, line_decode_table, macro_expressions
        )

        for func_name, func_body in self.functions.items():
            func_body = self.assemble_line_groups(
                func_body, line_decode_table, macro_expressions
            )
            if not func_body.startswith("\n"):
                func_body = "\n" + func_body

            if not func_body.endswith("\n"):
                func_body = func_body + "\n"

            is_struct = func_name.startswith("struct ")
            shader_code = shader_code.replace(
                format_function_name(func_name),
                f"{func_name} {{{func_body}}}" + (";" if is_struct else ""),
            )

        return shader_code

    @staticmethod
    def assemble_line_groups(
        code: list[CodeLineGroup],
        line_decode_table: list[str],
        macro_expressions: list[str],
    ):
        return "\n".join(
            group.assemble(line_decode_table, macro_expressions) for group in code
        )



from .type_aliases import FunctionName, ShaderFlags


class AllFlags:
    """
    This object stores all permutations of flags separately for each context (main code + functions)
    """

    main_flags: list[ShaderFlags]
    function_flags: dict[FunctionName, list[ShaderFlags]]

    def __init__(self):
        self.main_flags = []
        self.function_flags = {}



import re

from .processing import format_function_name
from .type_aliases import ShaderCode, FunctionName, ShaderFlags, ShaderLineIndex


class PermutationBase:
    code: ShaderCode
    flags: ShaderFlags

    def __init__(self):
        self.code = ""
        self.flags = {}


class FunctionPermutation(PermutationBase):
    is_struct: bool

    def __init__(self):
        super().__init__()
        self.is_struct = False


class ShaderPermutation(PermutationBase):
    functions: dict[FunctionName, FunctionPermutation]

    def __init__(self):
        super().__init__()
        self.functions = {}

    def extract_functions(self):
        """
        Extracts functions and structs from shader permutation code
        """
        re_func_start = re.compile(
            r"^[\s]*?([^#\s][\w]+)[\s]+([\w]+)[\s]*\(([^;]*?)\)[\s]*{",
            re.DOTALL | re.MULTILINE,
        )
        re_struct_start = re.compile(
            r"^[\s]*?struct[\s]+([\w]+)[\s]*{(.*?)};", re.DOTALL | re.MULTILINE
        )

        func_start = re.search(re_func_start, self.code)
        modified_code = ""

        # Extract functions.
        while func_start:
            groups = func_start.groups()
            args = groups[2].replace("\n", "")
            func_name = f"{groups[0]} {groups[1]}({args})"

            modified_code += self.code[: func_start.start()]

            bracket_balance = 1
            for i in range(func_start.end(), len(self.code)):
                c = self.code[i]
                if c == "{":
                    bracket_balance += 1
                elif c == "}":
                    bracket_balance -= 1

                if bracket_balance == 0:
                    func_content = self.code[func_start.end() : i]
                    self.code = self.code[i + 1 :]
                    break
            if bracket_balance:
                break

            function = FunctionPermutation()
            function.is_struct = False
            function.flags = self.flags
            function.code = func_content
            self.functions[func_name] = function

            modified_code += format_function_name(func_name) + "\n"

            func_start = re.search(re_func_start, self.code)

        self.code = modified_code + self.code

        # Extract structs.
        match: re.Match
        for match in re.finditer(re_struct_start, self.code):
            struct_name, struct_content = match.groups()
            struct_name = "struct " + struct_name

            struct = FunctionPermutation()
            struct.is_struct = True
            struct.flags = self.flags
            struct.code = struct_content
            self.functions[struct_name] = struct

            self.code = self.code.replace(
                match.group(), format_function_name(struct_name) + "\n"
            )


class EncodedUniqufiedPermutations:
    codes: list[list[ShaderLineIndex]]
    flags: list[list[ShaderFlags]]

    def __init__(self):
        self.codes = []
        self.flags = []



from ..platform import ShaderPlatform


class SupportedPlatforms:
    platforms: dict[ShaderPlatform, bool]

    def __init__(self) -> None:
        self.platforms = {platform: True for platform in ShaderPlatform}

    def _validate_bit_string(self, platforms_bit_string: str, length: int) -> str:
        if any(c not in "01" for c in platforms_bit_string):
            print(
                f"Warning: Invalid supported platforms bit field {platforms_bit_string}"
            )
            return "1" * length
        return platforms_bit_string

    def _format_bit_string(self, platforms_bit_string: str, length: int):
        # Truncate at the end if platform string is too long.
        platforms_bit_string = platforms_bit_string[:length]

        # In case the length is less than required, add leading zeros.
        platforms_bit_string = platforms_bit_string.zfill(length)

        return platforms_bit_string

    def parse_bit_string(self, platforms_bit_string: str, version: int):
        platform_list = ShaderPlatform.get_list(version)
        length = len(platform_list)

        platforms_bit_string = self._validate_bit_string(platforms_bit_string, length)
        platforms_bit_string = self._format_bit_string(platforms_bit_string, length)

        for platform, bit in zip(platform_list, platforms_bit_string):
            self.platforms[platform] = bit == "1"

        return self

    def get_bit_string(self, version: int) -> str:
        return "".join(
            "1" if self.platforms[p] else "0" for p in ShaderPlatform.get_list(version)
        )

    def load(self, json_data: dict[str, bool]):
        for key, value in json_data.items():
            self.platforms[ShaderPlatform[key]] = value

    def serialize(self, version: int):
        return {p.name: self.platforms[p] for p in ShaderPlatform.get_list(version)}



import os, pyjson5
from collections.abc import Callable
from typing import Any

from lazurite.material import Material
from lazurite.material.shader_pass.shader_definition import ShaderPlatform
from lazurite.compiler.macro_define import MacroDefine


class ProjectConfig:
    macros: list[MacroDefine]
    platforms: list[ShaderPlatform]
    merge_source: list[str]
    include_patterns: list[str]
    exclude_patterns: list[str]
    include_search_paths: list[str]

    def __init__(self) -> None:
        self.macros = []
        self.platforms = []
        self.merge_source = []
        self.include_patterns = ["*"]
        self.exclude_patterns = [".*", "_*"]
        self.include_search_paths = []

    def read_json_file(self, path: str, profiles: list[str]):
        if not os.path.isfile(path):
            return
        with open(path) as f:
            json_data = pyjson5.load(f)
        project_folder = os.path.split(path)[0]

        properties: list[tuple[list, str, Callable[[Any], list]]] = [
            (
                self.macros,
                "macros",
                lambda p: [MacroDefine.from_string(x) for x in p],
            ),
            (
                self.platforms,
                "platforms",
                lambda p: [ShaderPlatform[x] for x in p],
            ),
            (
                self.merge_source,
                "merge_source",
                lambda p: p,
            ),
            (
                self.include_patterns,
                "include_patterns",
                lambda p: p,
            ),
            (
                self.exclude_patterns,
                "exclude_patterns",
                lambda p: p,
            ),
            (
                self.include_search_paths,
                "include_search_paths",
                lambda p: [
                    os.path.normpath(os.path.join(project_folder, x)) for x in p
                ],
            ),
        ]
        updated_properties = {name: False for _, name, _ in properties}

        if "profiles" in json_data:
            json_profiles = json_data["profiles"]
            for profile in profiles:
                if profile not in json_profiles:
                    print(f'Warning: profile "{profile}" was not found!')
                    continue
                profile = json_profiles[profile]

                for property, property_name, value_getter in properties:
                    if property_name in profile:
                        if not updated_properties[property_name]:
                            property[:] = []
                            updated_properties[property_name] = True

                        values = value_getter(profile[property_name])
                        property.extend(
                            (item for item in values if item not in property)
                        )

        if "base_profile" in json_data:
            base_profile = json_data["base_profile"]

            for property, property_name, value_getter in properties:
                if (
                    property_name in base_profile
                    and not updated_properties[property_name]
                ):
                    property[:] = value_getter(base_profile[property_name])

        new_merge_source = []
        for merge_path in self.merge_source:
            merge_path = os.path.normpath(os.path.join(project_folder, merge_path))
            if (
                os.path.isfile(merge_path)
                and merge_path not in new_merge_source
                and merge_path.endswith((Material.EXTENSION, Material.JSON_EXTENSION))
            ):
                new_merge_source.append(merge_path)
            elif os.path.isdir(merge_path):
                for mat_dir in os.scandir(merge_path):
                    if (
                        mat_dir.is_file()
                        and mat_dir.path not in new_merge_source
                        and mat_dir.path.endswith(
                            (Material.EXTENSION, Material.JSON_EXTENSION)
                        )
                    ):
                        new_merge_source.append(mat_dir.path)
            else:
                print(f'Warning: merge path "{merge_path}" was not found')
        self.merge_source = new_merge_source



from dataclasses import dataclass

from .processing import postprocess_shader, preprocess_shader, strip_comments
from .expression_search import ExpressionSearchInput, expression_search
from .local_flag_definition import LocalFlagDeinition
from .type_aliases import ShaderCode, ShaderFlags
from .permutation import ShaderPermutation
from .encoded_shader import EncodedShader
from .expression_processing import (
    convert_to_sympy_expression,
    process_sympy_expressions,
    mark_approximated_results,
)


@dataclass
class InputVariant:
    """
    Decompiler input struct, containing code and flags.
    """

    flags: ShaderFlags
    code: ShaderCode


def restore_code(
    input_variants: list[InputVariant],
    remove_comments=True,
    process_shaders=False,
    search_timeout: float = 10,
) -> tuple[set[str], str]:
    """
    Attempts to restore original shader source, by combining variants while adding missing macros.
    """
    shader_permutations: list[ShaderPermutation] = []
    for variant in input_variants:
        code = variant.code

        if remove_comments:
            code = strip_comments(code)
        if process_shaders:
            code = preprocess_shader(code)

        permutation = ShaderPermutation()
        permutation.flags = variant.flags.copy()
        permutation.code = code
        permutation.extract_functions()

        shader_permutations.append(permutation)

    encoded_shader = EncodedShader(shader_permutations)
    diffed_shader = encoded_shader.diff()
    diffed_grouped_shader = diffed_shader.group_lines()

    local_flag_definition = LocalFlagDeinition.from_diffed_grouped_shader(
        diffed_grouped_shader
    )
    local_flag_definition.filter_and_bias_flags()
    all_flags = diffed_grouped_shader.gen_all_flags_list()

    expr_search_inputs = ExpressionSearchInput.from_diffed_grouped_shader(
        diffed_grouped_shader, local_flag_definition, all_flags
    )
    search_results = expression_search(expr_search_inputs, search_timeout)
    sympy_expressions = [
        convert_to_sympy_expression(res.token_list) for res in search_results
    ]
    macro_conditionals, used_macros = process_sympy_expressions(sympy_expressions)

    macro_conditionals = mark_approximated_results(
        macro_conditionals, search_results, expr_search_inputs
    )

    code = diffed_grouped_shader.assemble_code(
        macro_conditionals, encoded_shader.line_decode_table
    )

    if process_shaders:
        code = postprocess_shader(code)

    return used_macros, code



import sympy
import re

from lazurite import util

from .expression_search import (
    ExpressionSearchToken,
    JoinType,
    ExpressionSearchOutput,
    ExpressionSearchInput,
)


def convert_to_sympy_expression(tokens: list[ExpressionSearchToken]):
    expression: sympy.Symbol = sympy.false

    for token in tokens:
        if token.flag_name == "pass":
            macro = util.generate_pass_name_macro(token.flag_value)
        else:
            # TODO: refactor and/or better document the f_ thing
            if token.flag_name.startswith("f_"):
                flag_name = token.flag_name.removeprefix("f_")
                macro = util.generate_flag_name_macro(flag_name, token.flag_value)
            else:
                macro = util.format_definition_name(token.flag_name + token.flag_value)

        macro: sympy.Symbol = sympy.symbols(macro)

        if token.is_negative:
            macro = ~macro

        if token.join_type is JoinType.Initial:
            expression = macro
        elif token.join_type is JoinType.And:
            expression = expression & macro
        elif token.join_type is JoinType.Or:
            expression = expression | macro

    return expression


def _format_expression(expr: str):
    """
    Formats stringified sympy expression according to GLSL macro format.
    """
    expr = expr.replace("~", "!").replace("|", "||").replace("&", "&&")
    expr = re.sub(r"(\w+)", r"defined(\1)", expr)
    return expr


def process_sympy_expressions(expressions: list[sympy.Symbol]):
    """
    Processes a list of sympy expressions. The following logic is executed:
    - Simplifies sympy expressions
    - Converts simplified expressions into macro condition strings
    - Creates a set of macros that were referenced in simplified expressions
    """
    macros: set[str] = set()
    unique_results: list[str] = []
    unique_expressions = list(set(expressions))

    # Operate on unique expressions, in order to reduce compute workload (simplify_logic compute time grows exponentially)
    for expression in unique_expressions:
        expression = sympy.simplify_logic(expression, force=True)
        macro_condition = str(expression)
        macros.update(str(s) for s in expression.atoms())
        if len(expression.atoms()) == 1:
            # Special case handling of macro expressions with only 1 defined macro
            if macro_condition.startswith("~"):
                macro_condition = "#ifndef " + macro_condition.removeprefix("~")
            else:
                macro_condition = "#ifdef " + macro_condition
        else:
            macro_condition = "#if " + _format_expression(macro_condition)

        unique_results.append(macro_condition)

    return [
        unique_results[unique_expressions.index(expr)] for expr in expressions
    ], macros


def mark_approximated_results(
    macro_conditionals: list[str],
    search_results: list[ExpressionSearchOutput],
    search_inputs: list[ExpressionSearchInput],
):
    """
    This function adds the `// Approximation, matches X cases out of Y` comment to macro conditional expressions
    """
    output: list[str] = []
    for input, result, macro in zip(search_inputs, search_results, macro_conditionals):
        if result.score != len(input.flags):
            macro = f"// Approximation, matches {result.score} cases out of {len(input.flags)}\n{macro}"
        output.append(macro)

    return output



import re


def preprocess_shader(shader_code: str):
    """
    Pre-processes plain text shader code to convert it from GLSL to BGFX SC.\n
    Removes built-in `u_` uniforms, replaces `gl_FragColor` and `gl_FragData`,
    replaces attributes and varyings with `$input` and `$output`, removes macros,
    replaces samplers with BGFX AUTOREG macros, adds NUM_THREADS to compute shaders.
    """
    shader_code = re.sub(
        r"^uniform\s+\w+\s+u_[\w[\]]+;\n", "", shader_code, flags=re.MULTILINE
    )

    shader_code = re.sub(r"(\W)bgfx_FragColor(\W)", r"\1gl_FragColor\2", shader_code)
    shader_code = re.sub(r"(\W)bgfx_FragData(\W)", r"\1gl_FragData\2", shader_code)

    shader_code = re.sub(r"^out\s.+?;\n", "", shader_code, flags=re.MULTILINE)

    is_vertex_stage = bool(
        re.search(r"^#define varying out$", shader_code, flags=re.MULTILINE)
    )

    shader_code = re.sub(r"^#define\s.+?\n", "", shader_code, flags=re.MULTILINE)
    shader_code = re.sub(
        r"^#if\s.+?#endif\n", "", shader_code, flags=re.MULTILINE | re.DOTALL
    )
    shader_code = re.sub(r"^#extension\s.+?\n", "", shader_code, flags=re.MULTILINE)

    shader_code = re.sub(
        r"^[\s\w]*?varying\s.+? (\w+);$",
        r"$output \1" if is_vertex_stage else r"$input \1",
        shader_code,
        flags=re.MULTILINE,
    )
    shader_code = re.sub(
        r"^[\s\w]*?attribute\s.+? (\w+);$",
        r"$input \1",
        shader_code,
        flags=re.MULTILINE,
    )

    shader_code = re.sub(r"^#version\s.+?\n", "", shader_code)

    # TODO: Missing some samplers like `uniform lowp sampler2DArray s_WaterSurfaceDepthTextures`
    # (also missing from from bgfx_shader.sh)
    SAMPLERS = [
        (r"lowp sampler2D", r"SAMPLER2D"),
        (r"highp sampler2DMS", r"SAMPLER2DMS"),
        (r"highp sampler3D", r"SAMPLER3D"),
        (r"lowp samplerCube", r"SAMPLERCUBE"),
        (r"highp sampler2DShadow", r"SAMPLER2DSHADOW"),
        (r"highp sampler2D", r"SAMPLER2D_HIGHP"),
        (r"highp samplerCube", r"SAMPLERCUBE_HIGHP"),
        (r"highp sampler2DArray", r"SAMPLER2DARRAY"),
        (r"highp sampler2DMSArray", r"SAMPLER2DMSARRAY"),
        (r"highp samplerCubeArray", r"SAMPLERCUBEARRAY"),
        (r"highp sampler2DArrayShadow", r"SAMPLER2DARRAYSHADOW"),
        (r"highp isampler2D", r"ISAMPLER2D"),
        (r"highp usampler2D", r"USAMPLER2D"),
        (r"highp isampler3D", r"ISAMPLER3D"),
    ]

    for pattern, repl in SAMPLERS:
        pattern = r"^uniform " + pattern + r" (\w+);"
        repl = repl + r"_AUTOREG(\1);"
        shader_code = shader_code = re.sub(
            pattern, repl, shader_code, flags=re.MULTILINE
        )

    shader_code = shader_code = re.sub(
        r"^layout\(std430, .+?\) readonly buffer (\w+) { (\w+) .+? }",
        r"BUFFER_RO_AUTOREG(\1, \2);",
        shader_code,
        flags=re.MULTILINE,
    )
    shader_code = shader_code = re.sub(
        r"^layout\(std430, .+?\) writeonly buffer (\w+) { (\w+) .+? }",
        r"BUFFER_WR_AUTOREG(\1, \2);",
        shader_code,
        flags=re.MULTILINE,
    )
    shader_code = shader_code = re.sub(
        r"^layout\(std430, .+?\) buffer (\w+) { (\w+) .+? }",
        r"BUFFER_RW_AUTOREG(\1, \2)",
        shader_code,
        flags=re.MULTILINE,
    )

    for i, access in enumerate(("readonly ", "writeonly ", "")):
        for prefix in "iu":
            prefix = "u" if prefix == "u" else ""
            access_id = ("RO", "WR", "RW")[i]

            name = f"{prefix.upper()}IMAGE2D_{access_id}_AUTOREG"
            shader_code = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image2D (\w+)",
                name + r"(\2, \1)",
                shader_code,
                flags=re.MULTILINE,
            )

            name = f"{prefix.upper()}IMAGE2D_ARRAY_{access_id}_AUTOREG"
            shader_code = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image2DArray (\w+)",
                name + r"(\2, \1)",
                shader_code,
                flags=re.MULTILINE,
            )

            name = f"{prefix.upper()}IMAGE3D_{access_id}_AUTOREG"
            shader_code = re.sub(
                r"^layout\((.+?), .+?\) "
                + access
                + "uniform highp "
                + prefix
                + r"image3D (\w+)",
                name + r"(\2, \1)",
                shader_code,
                flags=re.MULTILINE,
            )

    shader_code = re.sub(
        r"^layout \(local_size_x = (\d+), local_size_y = (\d+), local_size_z = (\d+)\) in;",
        r"NUM_THREADS(\1, \2, \3)",
        shader_code,
        flags=re.MULTILINE,
    )

    return shader_code


def postprocess_shader(shader_code: str):
    """
    Post-processes plain text shader code to convert it from GLSL to BGFX SC.\n
    Merges `$input` and `$output` declarations together and adds `// Attention!`
    comment to potential array access and matrix multiplication operations.
    """
    shader_code = shader_code.splitlines(keepends=True)

    new_shader = []
    args = []
    line_type = 0  # 0 - none, 1 - input, 2 = output
    line_prefix = ""
    for line in shader_code:
        if line.startswith("$input "):
            current_line_type = 1
            line_prefix = "$input "
        elif line.startswith("$output "):
            current_line_type = 2
            line_prefix = "$output "
        else:
            current_line_type = 0

        if line_type:
            if line_type == current_line_type:
                args.append(line.removeprefix(line_prefix).removesuffix("\n"))
            else:
                new_shader.append(", ".join(args) + "\n")
        if not line_type or line_type != current_line_type:
            if current_line_type:
                args = [line.removesuffix("\n")]
            else:
                new_shader.append(line)

        line_type = current_line_type
    shader_code = new_shader

    for i, line in enumerate(shader_code):
        if ") * (" in line or "][" in line:
            line = line[:-1] + " // Attention!\n"
            shader_code[i] = line

    shader_code = "".join(shader_code)

    return shader_code


def format_function_name(name: str):
    """
    Formats function name such that it can be safely inserted in code and wouldn't conflict with valid GLSL.
    """
    return f"START_NAME|||{name}|||END_NAME"


def strip_comments(code: str):
    """
    Removes single line and multiline comments from GLSL code.
    """
    code = re.sub(r"//.*\n", "", code)
    code = re.sub(r"/\*.*?\*/", "", code, flags=re.DOTALL)
    code = re.sub(r"\n\n+", "\n", code)
    return code
